import { FormItem, Col, Input, TimeSelect, TimePicker, DatePicker, InputNumber, Switch, Slider, Rate } from 'element-ui'
import Select from '../select'
import CheckboxGroup from '../checkbox-group'
import RadioGroup from '../radio-group'
import Upload from '../upload'
import { isString, isFunction } from '../_util/is'
import { createPlaceholder } from './createPlaceholder'
import { cloneDeep, omit } from 'lodash'
import Spin from '../spin'
import RichText from '../rich-text'
import Cascader from '../cascader'
import { Fragment } from 'vue-fragment'

const FormItemProps = {
  ...omit(FormItem.props, 'rules'),
  /**
   * @description: 表单域的值
   * @param {*}
   * @return {*}
   */
  formModel: {
    type: Object,
    default: () => ({}),
  },
  /**
   * @description: 表单内元素
   * @param {*}
   * @return {*}
   */
  item: {
    type: Object,
    default: () => ({}),
  },
  /**
   * @description: 配合 label 属性使用，表示是否显示 label 后面的冒号
   * @param {*}
   * @return {*}
   */
  colon: {
    type: Boolean,
    default: true,
  },
}

export default {
  name: 'ZkFormItem',
  components: { Input, Select },
  inject: ['form'],
  props: FormItemProps,
  data() {
    return {
      clonedItem: cloneDeep(this.item),
      loading: false,
    }
  },
  computed: {
    modelValue: {
      get() {
        return this.formModel[this.clonedItem.field]
      },
      set(val) {
        this.$emit('change', val)
      },
    },
    /**
     * @description: 组件属性
     * @param {*}
     * @return {*}
     */
    componentProps() {
      const { formModel, clonedItem } = this
      const { componentProps = {}, component, label = '' } = clonedItem
      let compProps = componentProps
      // 通过函数指定组件属性
      if (isFunction(componentProps)) {
        compProps = componentProps({
          formModel,
          clonedItem,
        })
        compProps.placeholder ??= isString(component) ? createPlaceholder(component, label) : undefined
      }
      // 设置默认的placeholder
      if (isString(component)) {
        componentProps.placeholder ??= createPlaceholder(component, label)
      }
      let whiteList = []
      if (isString(component) && component === 'Upload') {
        whiteList = ['onError', 'onSuccess', 'onPreview', 'onRemove', 'onProgress', 'onChange', 'beforeUpload', 'beforeRemove', 'httpRequest', 'onExceed']
      }
      let events = {}
      const props = Object.keys(compProps).reduce((prev, key) => {
        if (/on([A-Z])/.test(key) && !whiteList.includes(key)) {
          const eventKey = key.replace(/on([A-Z])/, '$1').toLocaleLowerCase()
          events[eventKey] = compProps[key]
        } else if (!['class', 'style'].includes(key)) {
          if (isFunction(compProps[key])) {
            if (!whiteList.includes(key)) {
              prev[key] = compProps[key]({ formModel, clonedItem })
            } else {
              prev[key] = compProps[key]
            }
          } else {
            prev[key] = compProps[key]
          }
        }
        return prev
      }, {})
      return {
        events,
        props: props || {},
      }
    },
    /**
     * @description: 设置校验规则
     * @param {*}
     * @return {*}
     */
    rules() {
      const {
        rules: explicitRules = [], //给rules起个别名避免命名冲突
        component,
        label,
        required,
        dynamicRules,
      } = this.clonedItem
      // 动态验证规则优先
      if (isFunction(dynamicRules)) {
        const { item, formModel } = this
        return dynamicRules({ field: item.field, formModel, item })
      }
      let rules = cloneDeep(explicitRules)
      // 默认校验不通过的提示信息
      const defMsg = isString(component) ? createPlaceholder(component, label) : undefined
      // 用户没有指定校验规则但指定了required属性时添加默认校验规则
      if ((!rules || rules.length === 0) && required) {
        rules = [
          {
            required,
            message: defMsg,
            trigger: ['Input', 'InputNumber'].includes(component) ? 'blur' : 'change',
          },
        ]
      }
      return rules
    },
  },
  async mounted() {
    if (this.clonedItem.componentProps && this.clonedItem.componentProps.request) {
      const request = this.clonedItem.componentProps.request
      this.loading = true
      try {
        this.$set(this.clonedItem.componentProps, 'options', await request())
        this.$delete(this.clonedItem.componentProps, 'request')
      } finally {
        this.loading = false
      }
    }
  },
  methods: {
    handleActions(eName, e) {
      this.modelValue = e
      // 将事件继续向上抛给用户处理
      if (this.clonedItem.componentProps && this.clonedItem.componentProps[eName]) {
        if (!isFunction(this.clonedItem.componentProps[eName])) {
          throw new Error(`${eName} must be a callback function!`)
        }
        this.clonedItem.componentProps[eName](e, {
          formModel: this.formModel,
          item: this.clonedItem,
        })
      }
    },
  },
  render() {
    /**
     * @description: 渲染formItem内组件
     * @param {*}
     * @return {*}
     */
    const renderComponent = () => {
      // 组件是一个函数，执行函数返回的渲染结果
      if (typeof this.clonedItem.component === 'function') {
        return this.clonedItem.component({
          formModel: this.formModel,
          formRef: this.form,
          item: this.clonedItem,
        })
      } else {
        switch (this.clonedItem.component) {
          // *******输入框****** //
          case 'Input':
            return (
              <Input
                class={[this.clonedItem.componentProps?.class, 'zk-input']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                }}
                onInput={this.handleActions.bind(this, 'onInput')}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k](this.formModel)}
                    </Fragment>
                  ))}
              </Input>
            )
          // *******选择器****** //
          case 'Select':
            return (
              <Select
                class={[this.clonedItem.componentProps?.class, 'zk-select']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                    loading: this.loading,
                  },
                  on: { ...this.componentProps?.events },
                }}
                onChange={this.handleActions.bind(this, 'onChange')}>
                {Object.is(this.loading, true) && <Spin size='small' slot='empty' />}
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </Select>
            )
          // *******时间选择框****** //
          case 'TimeSelect':
            return (
              <TimeSelect
                class={[this.clonedItem.componentProps?.class, 'zk-select']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: {
                    ...this.componentProps?.events,
                  },
                  scopedSlots: {
                    ...this.item.componentSlots?.scopedSlots,
                  },
                }}
                // 源码中emit的是input事件
                onInput={this.handleActions.bind(this, 'onInput')}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </TimeSelect>
            )
          // *******时间选择框****** //
          case 'TimePicker':
            return (
              <TimePicker
                class={[this.clonedItem.componentProps?.class, 'zk-select']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: {
                    ...this.componentProps?.events,
                  },
                  scopedSlots: {
                    ...this.item.componentSlots?.scopedSlots,
                  },
                }}
                onInput={this.handleActions.bind(this, 'onInput')}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </TimePicker>
            )
          // *******日期选择框****** //
          case 'DatePicker':
            return (
              <DatePicker
                class={[this.clonedItem.componentProps?.class, 'jc-select']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                  scopedSlots: {
                    ...this.item.componentSlots?.scopedSlots,
                  },
                }}
                onInput={this.handleActions.bind(this, 'onInput')}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </DatePicker>
            )
          // *******数字输入框****** //
          case 'InputNumber':
            return (
              <InputNumber
                class={[this.clonedItem.componentProps?.class, 'zk-input-number']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                }}
                onChange={this.handleActions.bind(this, 'onChange')}></InputNumber>
            )
          // *******树形选择框****** //
          // case 'TreeSelect':
          //   return (
          //     <TreeSelect
          //       class={[this.clonedItem.componentProps?.class, 'zk-select']}
          //       style={this.clonedItem.componentProps?.style}
          //       {...{
          //         props: {
          //           value: this.modelValue,
          //           ...this.componentProps?.props,
          //         },
          //         on: { ...this.componentProps?.events },
          //         scopedSlots: {
          //           ...this.item.componentSlots?.scopedSlots,
          //         },
          //       }}
          //       onChange={this.handleActions.bind(this, 'onChange')}>
          //       {Object.is(this.loading, true) && <Spin size='small' slot='notFoundContent' />}
          //       {this.item.componentSlots?.slots &&
          //         Object.keys(this.item.componentSlots?.slots).map((k) => (
          //           <Fragment key={k} slot={k}>
          //             {this.item.componentSlots?.slots[k]()}
          //           </Fragment>
          //         ))}
          //     </TreeSelect>
          //   )
          // *******开关****** //
          case 'Switch':
            return (
              <Switch
                class={[this.clonedItem.componentProps?.class, 'zk-select']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                  scopedSlots: {
                    ...this.item.componentSlots?.scopedSlots,
                  },
                }}
                onChange={this.handleActions.bind(this, 'onChange')}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </Switch>
            )
          // *******复选框****** //
          case 'Checkbox':
            return (
              <CheckboxGroup
                class={[this.clonedItem.componentProps?.class, 'zk-select']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                }}
                onInput={this.handleActions.bind(this, 'onInput')}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </CheckboxGroup>
            )
          // *******级联选择器****** //
          case 'Cascader':
            return (
              <Cascader
                class={[this.clonedItem.componentProps?.class, 'zk-select']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  props: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                  scopedSlots: {
                    ...this.item.componentSlots?.scopedSlots,
                  },
                }}
                onChange={this.handleActions.bind(this, 'onChange')}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </Cascader>
            )
          // *******滑动选择器****** //
          case 'Slider':
            return (
              <Slider
                class={[this.clonedItem.componentProps?.class, 'zk-select']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                }}
                onInput={this.handleActions.bind(this, 'onInput')}></Slider>
            )
          // *******评分****** //
          case 'Rate':
            return (
              <Rate
                class={[this.clonedItem.componentProps?.class]}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                }}
                onChange={this.handleActions.bind(this, 'onChange')}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </Rate>
            )
          case 'Radio':
            return (
              <RadioGroup
                class={[this.clonedItem.componentProps?.class]}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    value: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                }}
                onInput={this.handleActions.bind(this, 'onInput')}></RadioGroup>
            )
          case 'Upload':
            return (
              <Upload
                class={[this.clonedItem.componentProps?.class, 'zk-upload']}
                style={this.clonedItem.componentProps?.style}
                {...{
                  attrs: {
                    ...this.componentProps?.props,
                    fileList: this.modelValue,
                  },
                }}>
                {this.item.componentSlots?.slots &&
                  Object.keys(this.item.componentSlots?.slots).map((k) => (
                    <Fragment key={k} slot={k}>
                      {this.item.componentSlots?.slots[k]()}
                    </Fragment>
                  ))}
              </Upload>
            )
          case 'RichText':
            return (
              <RichText
                class={[this.clonedItem.componentProps?.class]}
                style={this.clonedItem.componentProps?.style}
                {...{
                  props: {
                    ...this.componentProps?.props,
                    content: this.modelValue,
                  },
                  on: { ...this.componentProps?.events },
                }}
                onChange={this.handleActions.bind(this, 'onChange')}></RichText>
            )
          default:
            return null
        }
      }
    }
    // 组件hidden属性决定组件是否渲染
    const hidden = isFunction(this.clonedItem.hidden)
      ? this.clonedItem.hidden({
          formModel: this.formModel,
          item: this.clonedItem,
        })
      : this.clonedItem.hidden
    // 根据组件unwrap属性决定需不需要用FormItem包裹
    const formContent = !hidden ? (
      this.clonedItem.unWrap ? (
        <Col {...{ props: { ...this.clonedItem.colProps } }}> {renderComponent()}</Col>
      ) : (
        <Col {...{ props: { ...this.clonedItem.colProps } }}>
          <FormItem {...{ props: { ...this.clonedItem.formItemProps } }} label={this.clonedItem.label !== undefined && this.clonedItem.label !== null && this.colon ? this.clonedItem.label + ':' : this.clonedItem.label} rules={this.rules} prop={this.clonedItem.field} labelWidth={this.clonedItem.labelWidth}>
            {renderComponent()}
          </FormItem>
        </Col>
      )
    ) : null
    return formContent
  },
}
