import debounce from 'lodash/debounce'
const SpinProps = {
  spinning: {
    type: Boolean,
    default: false,
  },
  size: {
    type: String,
  },
  wrapperClassName: {
    type: String,
  },
  tip: {
    type: String,
  },
  delay: {
    type: Number,
  },
}
const prefixCls = 'zk-spin'

function shouldDelay(spinning, delay) {
  return !!spinning && !!delay && !isNaN(Number(delay))
}

export default {
  name: 'ZkSpin',
  props: SpinProps,
  data() {
    const { spinning, delay } = this
    const shouldBeDelayed = shouldDelay(spinning, delay)
    this.originalUpdateSpinning = this.updateSpinning
    this.debouncifyUpdateSpinning(this.$props)
    return {
      sSpinning: spinning && !shouldBeDelayed,
    }
  },
  mounted() {
    this.updateSpinning()
  },
  updated() {
    this.$nextTick(() => {
      this.debouncifyUpdateSpinning()
      this.updateSpinning()
    })
  },
  beforeDestroy() {
    this.cancelExistingSpin()
  },
  methods: {
    debouncifyUpdateSpinning(props) {
      const { delay } = props || this.$props
      if (delay) {
        this.cancelExistingSpin()
        this.updateSpinning = debounce(this.originalUpdateSpinning, delay)
      }
    },
    updateSpinning() {
      const { spinning, sSpinning } = this
      if (sSpinning !== spinning) {
        this.sSpinning = spinning
      }
    },
    cancelExistingSpin() {
      const { updateSpinning } = this
      if (updateSpinning && updateSpinning.cancel) {
        updateSpinning.cancel()
      }
    },
  },
  render() {
    const { size, tip, wrapperClassName, ...restProps } = this.$props
    const { sSpinning } = this
    const spinClassName = {
      [prefixCls]: true,
      [`${prefixCls}-mi`]: size === 'mini',
      [`${prefixCls}-sm`]: size === 'small',
      [`${prefixCls}-md`]: size === 'medium',
      [`${prefixCls}-spinning`]: sSpinning,
      [`${prefixCls}-show-text`]: !!tip,
    }

    const spinElement = (
      <div {...restProps} class={spinClassName}>
        {this.$slots.indicator || <i class='el-icon-loading' />}
        {tip ? <span class={`${prefixCls}-text`}>{tip}</span> : null}
      </div>
    )
    if (this.$slots.default) {
      const containerClassName = {
        [`${prefixCls}-container`]: true,
        [`${prefixCls}-blur`]: sSpinning,
      }
      return (
        <div class={[`${prefixCls}-nested-loading`, wrapperClassName]}>
          {sSpinning && spinElement}
          <div class={containerClassName} key='container'>
            {this.$slots.default}
          </div>
        </div>
      )
    }
    return spinElement
  },
}
