import { Pagination, Table, TableColumn } from 'element-ui'
import TableSettings from './table-settings'
import Spin from '../spin'
import Emitter from '../_util/emitter'
import Actions from './actions'
import QueryForm from './query'

const defaultPaginationProps = {
  layout: 'total, sizes, prev, pager, next, jumper',
  pageSizes: [10, 20, 30, 40, 50, 100],
  background: true,
  currentPage: 1,
  pageSize: 10,
  direction: 'right',
}

const TableProps = {
  ...Table.props,
  data: {
    type: [Function, Array],
    required: true,
  },
  showSetting: {
    type: Boolean,
    default: true,
  },
  columns: {
    type: Array,
    default: () => [],
  },
  pagination: {
    type: [Object, Boolean],
    default: () => defaultPaginationProps,
  },
  headerRowStyle: {
    type: [Function, Object],
    default: () => ({
      borderTopLeftRadius: '4px',
      borderTopRightRadius: '4px',
    }),
  },
  headerCellStyle: {
    type: [Function, Object],
    default: () => ({
      backgroundColor: '#F5F6FA',
      color: '#24292B',
      fontWeight: 600,
    }),
  },
  cellStyle: {
    color: '#515455',
    fontWeight: 400,
  },
  align: {
    type: String,
    default: 'left',
  },
  sortQuery: {
    type: Object,
    default: () => {},
  },
  filterQuery: {
    type: Object,
    default: () => {},
  },
  pageKey: {
    type: String,
    default: 'currentPage',
  },
  pageSizeKey: {
    type: String,
    default: 'pageSize',
  },
  // 是否自动调接口
  autoRequest: {
    type: Boolean,
    default: true,
  },
}

export default {
  componentName: 'ZkTable',
  name: 'ZkTable',
  components: { Table, TableColumn, Pagination },
  mixins: [Emitter],
  provide() {
    return {
      table: this,
      columns: this.columns.map((col) => ({ ...col, hideInTable: col.hideInTable || false })),
    }
  },
  props: TableProps,
  data() {
    return {
      listData: [],
      total: 0,
      sortParameters: {},
      pageSize: this.pagination.pageSize || 10,
      currentPage: this.pagination.currentPage || 1,
      loading: false,
      cols: [],
      query: {},
    }
  },
  watch: {
    columns: {
      handler() {
        this.initialize()
      },
      immediate: true,
    },
    sortQuery: {
      handler() {
        this.reloadData(true)
      },
      deep: true,
    },
    filterQuery: {
      handler() {
        this.reloadData(true)
      },
      deep: true,
    },
  },
  created() {
    // 判断数据源的类型
    if (typeof this.data === 'function') {
      // 需要自动加载时才会自定执行这个function
      if (this.autoRequest) {
        this.loadData()
      }
    } else {
      this.listData = this.pagination !== false ? this.data.slice((this.currentPage - 1) * this.pageSize, this.pageSize) : this.data
      this.total = this.data.length
    }
  },
  mounted() {
    // 收到隐藏或显示列事件
    this.$on('change-hide', ({ hide, index }) => {
      // 直接通过下标操作数组无法触发响应式的依赖收集，使用Set可以触发
      this.$set(this.cols, index, { ...this.cols[index], hideInTable: hide })
    })
    // 收到更改列排序事件
    this.$on('change-sort', ({ newIndex, oldIndex }) => {
      this.cols.splice(newIndex, 0, this.cols.splice(oldIndex, 1)[0])
    })
    // 收到重置事件
    this.$on('reset', () => {
      this.initialize()
    })
    this.$nextTick(() => {
      const functions = Table.methods
      // 接管Table的方法
      for (const item in functions) {
        this[item] = function () {
          this.$refs.table[item].apply(this, arguments)
        }
      }
    })
  },
  methods: {
    initialize() {
      // 将源数据拷贝一份，防止后续操作污染源数据引起重置不可逆
      const cols = [...this.columns]
      // 列数据当中有筛选且filters是一个function时，执行这个function获取数据
      cols.forEach(async (col) => {
        if (col.filters && typeof col.filters === 'function') {
          col.filters = await col.filters()
        }
      })
      // 必要的，确保筛选数据已经拿到再渲染
      this.$nextTick(() => {
        this.cols = cols
      })
    },
    /**
     * @description: 获取远程数据源
     * @param {*}
     * @return {*}
     */
    loadData() {
      this.loading = true
      const { pageKey, pageSizeKey, query } = this
      let parameters = {}
      if (this.pagination) {
        parameters[pageKey] = this.currentPage
        parameters[pageSizeKey] = this.pageSize
      }

      // 是否有排序
      typeof this.sortQuery === 'object' && Object.keys(this.sortQuery) && (parameters = { ...parameters, ...this.sortQuery })

      // 是否有筛选
      typeof this.filterQuery === 'object' && Object.keys(this.filterQuery) && (parameters = { ...parameters, ...this.filterQuery })

      parameters = { ...parameters, ...query }
      const result = this.data(parameters)
      result
        .then((res) => {
          // 数据删除后当前页为0时请求上一页数据
          if (Array.isArray(res)) {
            this.listData = res
          } else {
            if (res.data.length === 0 && this.pagination && this.currentPage > 1) {
              this.currentPage--
              this.loadData()
              return
            }
            // 处理分页信息
            const { total } = res
            this.total = Number(total)
            this.listData = res.data
          }
        })
        .finally(() => {
          this.loading = false
        })
    },
    /**
     * @description: 刷新数据
     * @param {*} bool 页码是否重置到第一页
     * @return {*}
     */
    reloadData(bool = false) {
      if (bool) {
        this.currentPage = 1
      }
      if (typeof this.data === 'function') {
        this.loadData()
      } else {
        this.loading = true
        this.listData = this.pagination !== false ? this.data.slice((this.currentPage - 1) * this.pageSize, this.pageSize) : this.data
        // 添加延迟
        setTimeout(() => {
          this.loading = false
        }, 300)
      }
    },
    queryAction(query) {
      this.query = query
      this.reloadData(true)
    },
    resetAction() {
      this.query = {}
      this.reloadData(true)
    },
  },
  render() {
    const { cols, showSetting, pagination, listData, pageSize, currentPage, total, align } = this
    // 组装Table的props
    const tempProps = Object.keys(Table.props).reduce((prev, cur) => {
      prev[cur] = this[cur]
      return prev
    }, {})
    const tableProps = {
      props: { ...tempProps, data: listData },
      on: {
        ...this.$listeners,
      },
    }
    // pagination相关属性及事件
    const paginationProps = {
      props: {
        ...defaultPaginationProps,
        ...pagination,
        pageSize: pageSize,
        total,
        currentPage: currentPage,
      },
      on: {
        'size-change': (size) => {
          // 更改size
          this.pageSize = size
          // 重新获取数据
          this.reloadData(true)
        },
        'current-change': (page) => {
          this.currentPage = page
          // 重新获取数据
          this.reloadData()
        },
      },
    }
    /**
     * @description: 渲染列
     * @param {*} columns 列数据
     * @param {*} align 对齐方式
     * @return {*}
     */
    const renderTableColumns = (columns, align) => {
      return columns.map((column, index) => {
        // 组装TableColumn的props
        const columnProps = Object.keys(TableColumn.props).reduce(
          (prev, cur) => {
            prev[cur] = column[cur]
            return prev
          },
          { align }
        )
        const scopedSlots =
          column.scopedSlots ||
          (column.prop === '$$action'
            ? {
                default: (slot) => {
                  return <Actions column={column} actions={column.actions?.call(undefined, slot)}></Actions>
                },
              }
            : null)
        // 必须指定key，否则会引起错误渲染
        return (
          <el-table-column
            {...{
              props: {
                ...columnProps,
                align,
              },
              key: column.prop + column.label + index,
              scopedSlots,
            }}>
            {column.children && column.children.length > 0 && renderTableColumns(column.children, align)}
          </el-table-column>
        )
      })
    }

    return (
      /**
       * @description: 用Spin包裹Table可以给Table一个局部刷新的效果
       * @param {*}
       * @return {*}
       */
      <Spin spinning={this.loading}>
        <QueryForm columns={cols} onQuery={this.queryAction} onReset={this.resetAction}></QueryForm>
        <div class={['handler-area', !this.$slots.actions && 'right-only']}>
          {this.$slots.actions && <div class='handler-left'>{this.$slots.actions}</div>}
          {showSetting && <TableSettings class='handler-right'></TableSettings>}
        </div>
        <el-table {...tableProps} ref='table'>
          {renderTableColumns(
            cols.filter((col) => !col.hideInTable),
            align
          )}
        </el-table>
        {pagination && (
          <div class='pagination' style={{ 'text-align': paginationProps.props.direction }}>
            <Pagination {...paginationProps} />
          </div>
        )}
      </Spin>
    )
  },
}
